/*
Copyright (C) 2024 uc-101010.com

Permission to use, copy, modify, and/or distribute this software for any purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/


package main

import (
	"image"
	"image/color"
	"image/gif"
	"os"
)

func main() {
	// Declare variables for the Lorenz attractor
	var x1, y1, z1 float64
	var x2, y2, z2 float64
	var sigma float64 = 10.0
	var beta float64 = 8.0 / 3.0
	var rho float64 = 28.0
	var dt float64 = 0.001

	// Set the initial point for the attractor
	x1, y1, z1 = 99, .10, 0

	// Set the number of iterations to simulate
	iterations_to_simulate := 1000000

	// Set the exposure window (number of iterations per frame)
	const exposure_window = 10000

	// Set the size of the output image in pixels
	size := 200

	// Set the delay between frames in 10ms units
	delay := 30

	// Define the color palette (black and white)
	var palette = []color.Color{color.Black, color.White}
	blackIndex := 1 // Index of black color in the palette

	// Create GIF animations for each projection (xy, xz, yz)
	anim_xy := gif.GIF{}
	anim_xz := gif.GIF{}
	anim_yz := gif.GIF{}

	// Create image rectangles for each projection
	rect := image.Rect(0, 0, 2*size+1, 2*size+1)
	img_xy := image.NewPaletted(rect, palette)
	img_xz := image.NewPaletted(rect, palette)
	img_yz := image.NewPaletted(rect, palette)

	// Simulate the Lorenz attractor
	for i := 0; i < iterations_to_simulate; i++ {
		// Calculate the next point delta using the Lorenz equations
		x2 = dt * (sigma * (y1 - x1))
		y2 = dt * (rho*x1 - y1 - x1*z1)
		z2 = dt * (x1*y1 - beta*z1)

		// Reset the image for each exposure window
		if i%exposure_window == 0 {
			img_xy = image.NewPaletted(rect, palette)
			img_xz = image.NewPaletted(rect, palette)
			img_yz = image.NewPaletted(rect, palette)
		}

		// Set the scale factor for plotting points
		const scale_factor = 7

		// Plot the point on each projection
		img_xy.SetColorIndex(size+int(x1*scale_factor), size-int(y1*scale_factor),
			uint8(blackIndex))
		img_xz.SetColorIndex(size+int(x1*scale_factor), 2*size-int(z1*scale_factor),
			uint8(blackIndex))
		img_yz.SetColorIndex(size+int(y1*scale_factor), 2*size-int(z1*scale_factor),
			uint8(blackIndex))

		// Append the image to the animation for each exposure window
		if i%exposure_window == 0 {
			anim_xy.Delay = append(anim_xy.Delay, delay)
			anim_xy.Image = append(anim_xy.Image, img_xy)
			anim_xz.Delay = append(anim_xz.Delay, delay)
			anim_xz.Image = append(anim_xz.Image, img_xz)
			anim_yz.Delay = append(anim_yz.Delay, delay)
			anim_yz.Image = append(anim_yz.Image, img_yz)
		}

		// Update the current point by adding the delta (Euler method)
		x1 = x1 + x2
		y1 = y1 + y2
		z1 = z1 + z2
	}

	// Create output files for each projection
	f_xy, _ := os.Create("xy.gif")
	f_xz, _ := os.Create("xz.gif")
	f_yz, _ := os.Create("yz.gif")

	// Close the files when done
	defer f_xy.Close()
	defer f_xz.Close()
	defer f_yz.Close()

	// Create channels for signaling when encoding is done
	done_xy, done_xz, done_yz := make(chan int), make(chan int), make(chan int)

	// Encode the animations in separate goroutines
	go func() {
		gif.EncodeAll(f_xy, &anim_xy)
		done_xy <- 1 // Signal the main goroutine when done
	}()
	go func() {
		gif.EncodeAll(f_xz, &anim_xz)
		done_xz <- 1 // Signal the main goroutine when done
	}()
	go func() {
		gif.EncodeAll(f_yz, &anim_yz)
		done_yz <- 1 // Signal the main goroutine when done
	}()

	// Wait for background goroutines to finish
	<-done_xy
	<-done_xz
	<-done_yz
}
